from datetime import timedelta
from typing import Literal

from prefect import flow, task
from prefect.tasks import task_input_hash
from prefect.utilities.annotations import quote

import raggy
from raggy.documents import Document
from raggy.loaders.base import Loader
from raggy.loaders.web import SitemapLoader
from raggy.vectorstores.chroma import Chroma


def html_parser(html: str) -> str:
    import trafilatura

    trafilatura_config = trafilatura.settings.use_config()
    # disable signal, so it can run in a worker thread
    # https://github.com/adbar/trafilatura/issues/202
    trafilatura_config.set("DEFAULT", "EXTRACTION_TIMEOUT", "0")
    return trafilatura.extract(html, config=trafilatura_config)


raggy.settings.html_parser = html_parser

prefect_website_loaders = [
    SitemapLoader(
        urls=["https://docs.prefect.io/sitemap.xml", "https://prefect.io/sitemap.xml"],
        exclude=["api-ref", "/events/"],
    )
]


@task(
    retries=2,
    retry_delay_seconds=[3, 60],
    cache_key_fn=task_input_hash,
    cache_expiration=timedelta(days=1),
    task_run_name="Run {loader.__class__.__name__}",
    persist_result=True,
    # refresh_cache=True,
)
async def run_loader(loader: Loader) -> list[Document]:
    return await loader.load()


@flow(name="Update Knowledge", log_prints=True)
async def refresh_chroma(
    collection_name: str = "default",
    chroma_client_type: str = "base",
    mode: Literal["upsert", "reset"] = "upsert",
):
    """Flow updating vectorstore with info from the Prefect community."""
    documents = [
        doc
        for future in await run_loader.map(quote(prefect_website_loaders))
        for doc in await future.result()
    ]

    print(f"Loaded {len(documents)} documents from the Prefect community.")

    async with Chroma(
        collection_name=collection_name, client_type=chroma_client_type
    ) as chroma:
        if mode == "reset":
            await chroma.reset_collection()
            docs = await chroma.add(documents)
        elif mode == "upsert":
            docs = await chroma.upsert(documents)
        else:
            raise ValueError(f"Unknown mode: {mode!r} (expected 'upsert' or 'reset')")

        print(f"Added {len(docs)} documents to the {collection_name} collection.")


if __name__ == "__main__":
    import asyncio

    asyncio.run(
        refresh_chroma(
            collection_name="testing", chroma_client_type="base", mode="reset"
        )
    )



from typing import Iterable

import turbopuffer as tpuf
from pydantic import (
    Field,
    SecretStr,
    model_validator,
)
from pydantic_settings import BaseSettings, SettingsConfigDict
from turbopuffer.vectors import VectorResult

from raggy.documents import Document
from raggy.utilities.asyncutils import run_sync_in_worker_thread
from raggy.utilities.embeddings import create_openai_embeddings
from raggy.utilities.text import slice_tokens
from raggy.vectorstores.base import Vectorstore


class TurboPufferSettings(BaseSettings):
    """Settings for the TurboPuffer vectorstore."""

    model_config = SettingsConfigDict(
        env_prefix="RAGGY_TURBOPUFFER_",
        env_file=("~/.raggy/.env", ".env"),
        arbitrary_types_allowed=True,
        extra="ignore",
    )

    api_key: SecretStr
    default_namespace: str = "raggy"

    @model_validator(mode="after")
    def set_api_key(self):
        if not tpuf.api_key and self.api_key:
            tpuf.api_key = self.api_key.get_secret_value()


tpuf_settings = TurboPufferSettings()


class TurboPuffer(Vectorstore):
    """Wrapper for turbopuffer.Namespace as an async context manager.

    Attributes:
        namespace: The namespace to use for the TurboPuffer instance.

    Examples:
        Upsert documents to a namespace:
        ```python
        from raggy.documents import Document
        from raggy.vectorstores.tpuf import TurboPuffer

        async with TurboPuffer() as tpuf: # default namespace is "raggy"
            await tpuf.upsert(documents=[Document(id="1", text="Hello, world!")])
        ```

        Query a namespace:
        ```python
        from raggy.vectorstores.tpuf import TurboPuffer

        async with TurboPuffer() as tpuf:
            result = await tpuf.query(text="Hello, world!")
            print(result)
        ```
    """

    namespace: str = Field(default_factory=lambda: tpuf_settings.default_namespace)

    @property
    def ns(self):
        return tpuf.Namespace(self.namespace)

    async def upsert(
        self,
        documents: Iterable[Document] | None = None,
        ids: list[str] | list[int] | None = None,
        vectors: list[list[float]] | None = None,
        attributes: dict | None = None,
    ):
        attributes = attributes or {}

        if documents is None and vectors is None:
            raise ValueError("Either `documents` or `vectors` must be provided.")

        if documents:
            ids = [document.id for document in documents]
            vectors = await create_openai_embeddings(
                [document.text for document in documents]
            )
            if attributes.get("text"):
                raise ValueError(
                    "The `text` attribute is reserved and cannot be used as a custom attribute."
                )
            attributes |= {"text": [document.text for document in documents]}

        await run_sync_in_worker_thread(
            self.ns.upsert, ids=ids, vectors=vectors, attributes=attributes
        )

    async def query(
        self,
        text: str | None = None,
        vector: list[float] | None = None,
        top_k: int = 10,
        distance_metric: str = "cosine_distance",
        filters: dict | None = None,
        include_attributes: list[str] | None = None,
        include_vectors: bool = False,
    ) -> VectorResult:
        if text:
            vector = await create_openai_embeddings(text)
        else:
            if vector is None:
                raise ValueError("Either `text` or `vector` must be provided.")

        return await run_sync_in_worker_thread(
            self.ns.query,
            vector=vector,
            top_k=top_k,
            distance_metric=distance_metric,
            filters=filters,
            include_attributes=include_attributes or ["text"],
            include_vectors=include_vectors,
        )

    async def delete(self, ids: str | int | list[str] | list[int]):
        await run_sync_in_worker_thread(self.ns.delete, ids=ids)

    async def reset(self):
        try:
            await run_sync_in_worker_thread(self.ns.delete_all)
        except tpuf.APIError as e:
            if e.status_code == 404:
                self.logger.debug_kv("404", "Namespace already empty.")
            else:
                raise

    async def ok(self) -> bool:
        try:
            return await run_sync_in_worker_thread(self.ns.exists)
        except tpuf.APIError as e:
            if e.status_code == 404:
                self.logger.debug_kv("404", "Namespace does not exist.")
                return False
            raise


async def query_namespace(
    query_text: str,
    filters: dict | None = None,
    namespace: str = "raggy",
    top_k: int = 10,
    max_tokens: int = 500,
) -> str:
    """Query a TurboPuffer namespace.

    Args:
        query_text: The text to query for.
        filters: Filters to apply to the query.
        namespace: The namespace to query.
        top_k: The number of results to return.

    Examples:
        Basic Usage of `query_namespace`
        ```python
        from raggy.vectorstores.tpuf import query_namespace

        print(await query_namespace("How to create a flow in Prefect?"))
        ```

        Using `filters` with `query_namespace`
        ```python
        from raggy.vectorstores.tpuf import query_namespace

        filters={
            'id': [['In', [1, 2, 3]]],
            'key1': [['Eq', 'one']],
            'filename': [['Or', [['Glob', '**.md'], ['Glob', '**.py']]], ['NotGlob', '/migrations/**']]
        }

        print(await query_namespace("How to create a flow in Prefect?", filters=filters))
        ```
    """
    async with TurboPuffer(namespace=namespace) as tpuf:
        vector_result = await tpuf.query(
            text=query_text,
            filters=filters,
            top_k=top_k,
        )

        concatenated_result = "\n".join(
            row.attributes["text"] for row in vector_result.data
        )

        return slice_tokens(concatenated_result, max_tokens)



from pydantic import BaseModel, ConfigDict, PrivateAttr

from raggy.utilities.logging import RaggyLogger, get_logger


class Vectorstore(BaseModel):
    """Base class for vectorstores.

    Allows for easy logging and async context management.

    Attributes:
        _in_context: Whether the vectorstore is currently in an async context.

    Example:
        Basic Usage of `Vectorstore`
        ```python
        from raggy.vectorstores.base import Vectorstore

        class MyVectorstore(Vectorstore):
            pass

        async with MyVectorstore() as vectorstore:
            ...
        ```
    """

    model_config = ConfigDict(arbitrary_types_allowed=True)

    _in_context: bool = PrivateAttr(False)

    @property
    def logger(self) -> RaggyLogger:
        return get_logger(self.__class__.__name__)

    async def __aenter__(self):
        self._in_context = True
        return self

    async def __aexit__(self, exc_type, exc_value, traceback):
        self._in_context = False



import asyncio
import re
from urllib.parse import urljoin

from bs4 import BeautifulSoup
from fake_useragent import UserAgent
from httpx import AsyncClient, Response
from pydantic import Field

import raggy
from raggy.documents import Document, document_to_excerpts
from raggy.loaders.base import Loader, MultiLoader
from raggy.utilities.collections import batched

user_agent = UserAgent()

URL_CONCURRENCY = asyncio.Semaphore(30)


def ensure_http(url):
    if not url.startswith(("http://", "https://")):
        return "http://" + url
    return url


async def sitemap_search(sitemap_url) -> list[str]:
    async with AsyncClient() as client:
        response = await client.get(sitemap_url, follow_redirects=True)
        response.raise_for_status()

    soup = BeautifulSoup(response.content, "xml")

    return [loc.text for loc in soup.find_all("loc")]


class WebLoader(Loader):
    document_type: str = "web page"
    headers: dict = Field(default_factory=dict, repr=False)

    async def get_headers(self) -> dict:
        return {"User-Agent": user_agent.random, **self.headers}


class URLLoader(WebLoader):
    """
    Given a list of URLs, loads whatever it finds there.

    Attributes:
        urls: The URLs to load from.
    """

    source_type: str = "url"

    urls: list[str] = Field(default_factory=list)

    async def load(self) -> list[Document]:
        headers = await self.get_headers()
        async with AsyncClient(
            headers=headers, timeout=30, follow_redirects=True
        ) as client:
            documents = await asyncio.gather(
                *[self.load_url(u, client) for u in self.urls], return_exceptions=True
            )
        final_documents = []
        for d in documents:
            if isinstance(d, Exception):
                self.logger.error(d)
            elif d is not None:
                final_documents.extend(await document_to_excerpts(d))  # type: ignore
        return final_documents

    async def load_url(self, url, client) -> Document | None:
        async with URL_CONCURRENCY:
            response = await client.get(url, follow_redirects=True)

        if not response.status_code == 200:
            self.logger.warning_style(
                f"Received status {response.status_code} from {url}", "red"
            )

        # check for a meta refresh redirect in the response content
        soup = BeautifulSoup(response.text, "html.parser")
        meta_refresh = soup.find(
            "meta", attrs={"http-equiv": re.compile(r"refresh", re.I)}
        )
        if meta_refresh:
            refresh_content = meta_refresh.get("content")
            redirect_url_match = re.search(r"url=([\S]+)", refresh_content, re.I)
            if redirect_url_match:
                redirect_url = redirect_url_match.group(1)
                # join base url with relative url
                redirect_url = urljoin(str(response.url), redirect_url)
                # Now ensure the URL includes the protocol
                redirect_url = ensure_http(redirect_url)
                response = await client.get(redirect_url, follow_redirects=True)

        document = await self.response_to_document(response)
        if document:
            self.logger.info(f"Loaded document from {url}")
        else:
            self.logger.warning_style(f"Could not load document from {url}", "red")
        return document

    async def response_to_document(self, response: Response) -> Document:
        return Document(
            text=await self.get_document_text(response),
            metadata=dict(
                link=str(response.url),
                source=self.source_type,
                document_type=self.document_type,
            ),
        )

    async def get_document_text(self, response: Response) -> str:
        return response.text


class HTMLLoader(URLLoader):
    """
    A loader that loads HTML, optionally converting it to markdown or stripping tags
    """

    async def get_document_text(self, response: Response) -> str:
        text = await super().get_document_text(response)
        return raggy.settings.html_parser(text)


class SitemapLoader(URLLoader):
    """A loader that loads URLs from a sitemap.

    Attributes:
        include: A list of strings or regular expressions. Only URLs that match one of these will be included.
        exclude: A list of strings or regular expressions. URLs that match one of these will be excluded.
        url_loader: The loader to use for loading the URLs.

    Examples:
        Load all URLs from a sitemap:
        ```python
        from raggy.loaders.web import SitemapLoader
        loader = SitemapLoader(urls=["https://askmarvin.ai/sitemap.xml"])
        documents = await loader.load()
        print(documents)
        ```
    """

    include: list[str | re.Pattern] = Field(default_factory=list)
    exclude: list[str | re.Pattern] = Field(default_factory=list)
    url_loader: URLLoader = Field(default_factory=HTMLLoader)

    async def _get_loader(self) -> Loader:
        urls = await asyncio.gather(*[self.load_sitemap(url) for url in self.urls])
        return MultiLoader(
            loaders=[
                type(self.url_loader)(urls=url_batch, headers=await self.get_headers())
                for url_batch in batched([u for url_list in urls for u in url_list], 10)
            ]
        )

    async def load(self) -> list[Document]:
        loader = await self._get_loader()
        return await loader.load()

    async def load_and_store(self, **kwargs) -> list[str]:
        loader = await self._get_loader()
        return await loader.load_and_store(**kwargs)

    async def load_sitemap(self, url: str) -> list[str]:
        def is_included(url: str) -> bool:
            if not self.include:
                return True

            return any(
                (isinstance(i, str) and i in url)
                or (isinstance(i, re.Pattern) and re.search(i, url))
                for i in self.include
            )

        def is_excluded(url: str) -> bool:
            return any(
                (isinstance(e, str) and e in url)
                or (isinstance(e, re.Pattern) and re.search(e, url))
                for e in self.exclude
            )

        return [
            url
            for url in await sitemap_search(url)
            if is_included(url) and not is_excluded(url)
        ]



import logging
import subprocess

log = logging.getLogger("mkdocs")


def on_pre_build(config, **kwargs):
    """Add a custom route to the server."""
    try:
        subprocess.run(
            [
                "npx",
                "tailwindcss",
                "-i",
                "./docs/overrides/tailwind.css",
                "-o",
                "./docs/static/css/tailwind.css",
            ]
        )
    except Exception:
        log.error("You need to install tailwindcss using npx install tailwindcss")



from typing import Callable

from bs4 import BeautifulSoup
from pydantic import Field, field_validator
from pydantic_settings import BaseSettings, SettingsConfigDict


def default_html_parser(html: str) -> str:
    """The default HTML parser. This uses `bs4`'s `html.parser`, which is not very good.
    Like, at all.

    In fact it's really bad. You should definitely set `raggy.settings.html_parser` to a
    `Callable[[str], str]` that parses HTML well.

    Args:
        html: The HTML to parse.

    Returns:
        The parsed HTML.
    """
    from raggy.utilities.logging import get_logger

    get_logger().warning_kv(
        "USING DEFAULT HTML PARSER",
        (
            "BeautifulSoup's html.parser is the default parser and is not very good. "
            "Consider setting `raggy.settings.html_parser` to a `Callable[[str], str]` that parses HTML well."
        ),
        "red",
    )
    return BeautifulSoup(html, "html.parser").get_text()


class Settings(BaseSettings):
    """The settings for Raggy.

    Attributes:
        html_parser: The function to use for parsing HTML.
        log_level: The log level to use.
        log_verbose: Whether to log verbose messages.
        openai_chat_completions_model: The OpenAI model to use for chat completions.
        openai_embeddings_model: The OpenAI model to use for creating embeddings.

    """

    model_config = SettingsConfigDict(
        env_prefix="RAGGY_",
        env_file=("~/.raggy/.env", ".env"),
        extra="allow",
        validate_assignment=True,
    )

    html_parser: Callable[[str], str] = default_html_parser

    log_level: str = Field(
        default="INFO",
        description="The log level to use.",
    )

    log_verbose: bool = Field(
        default=False,
        description=(
            "Whether to log verbose messages, such as full API requests and responses."
        ),
    )
    openai_chat_completions_model: str = Field(
        default="gpt-3.5-turbo",
        description="The OpenAI model to use for chat completions.",
    )

    openai_embeddings_model: str = Field(
        default="text-embedding-3-small",
        description="The OpenAI model to use for creating embeddings.",
    )

    @field_validator("log_level", mode="after")
    @classmethod
    def set_log_level(cls, v):
        from raggy.utilities.logging import setup_logging

        setup_logging(level=v)
        return v


settings = Settings()



import uuid


def generate_prefixed_uuid(prefix: str) -> str:
    """Generate a UUID string with the given prefix.

    Args:
        prefix: The prefix to use for the UUID

    Returns:
        A UUID string with the given prefix

    Raises:
        ValueError: If the prefix contains an underscore

    Example:
        Generate a UUID with the prefix "my_prefix"
        ```python
        from raggy.utilities.ids import generate_prefixed_uuid

        uuid = generate_prefixed_uuid("schleeb") # 'schleeb_6be20040-b7e0-4990-b271-394221584a59'
        ```

    """
    if "_" in prefix:
        raise ValueError("Prefix must not contain underscores.")
    return f"{prefix}_{uuid.uuid4()}"



import asyncio
import inspect
from functools import partial
from typing import Annotated, Iterable

from jinja2 import Environment, Template
from pydantic import BaseModel, ConfigDict, Field, model_validator

from raggy.utilities.collections import distinct
from raggy.utilities.ids import generate_prefixed_uuid
from raggy.utilities.text import count_tokens, extract_keywords, hash_text, split_text

jinja_env = Environment(enable_async=True)


class DocumentMetadata(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, extra="allow")

    title: str | None = Field(default=None)
    link: str | None = Field(default=None)


class Document(BaseModel):
    """A source of information that is storable & searchable.

    Anything that can be represented as text can be stored as a document:
    web pages, git repos / issues, PDFs, and or just plain text files.
    """

    model_config = ConfigDict(arbitrary_types_allowed=True, extra="allow")

    id: str = Field(default_factory=partial(generate_prefixed_uuid, "doc"))
    text: str = Field(..., description="Document text content.")

    embedding: list[float] | None = Field(default=None)
    metadata: DocumentMetadata = Field(default_factory=DocumentMetadata)

    tokens: int | None = Field(default=None)
    keywords: list[str] = Field(default_factory=list)

    @property
    def hash(self):
        return hash_text(self.text)

    @model_validator(mode="after")
    def validate_tokens(self):
        if self.tokens is None:
            self.tokens = count_tokens(self.text)
        return self

    def __hash__(self) -> int:
        return hash(self.hash)


EXCERPT_TEMPLATE = jinja_env.from_string(
    inspect.cleandoc(
        """The following is an excerpt from a document
        {% if document.metadata %}\n\n# Document metadata
        {{ document.metadata }}
        {% endif %}
        {% if document.keywords %}
        # Document keywords
        {{ document.keywords }}
        {% endif %}
        # Excerpt content: {{ excerpt_text }}
        """
    )
)


async def document_to_excerpts(
    document: Document,
    excerpt_template: Template | None = None,
    chunk_tokens: int = 300,
    overlap: Annotated[float, Field(strict=True, ge=0, le=1)] = 0.1,
    **extra_template_kwargs,
) -> list[Document]:
    """
    Create document excerpts by chunking the document text into regularly-sized
    chunks and adding a "minimap" directory to the top (if document is markdown).

    Args:
        excerpt_template: A jinja2 template to use for rendering the excerpt.
        chunk_tokens: The number of tokens to include in each excerpt.
        overlap: The fraction of overlap between each excerpt.

    """
    if not excerpt_template:
        excerpt_template = EXCERPT_TEMPLATE

    text_chunks: list[str] = split_text(
        text=document.text,
        chunk_size=chunk_tokens,
        chunk_overlap=overlap,
    )

    return await asyncio.gather(
        *[
            _create_excerpt(
                document=document,
                text=text,
                index=i,
                excerpt_template=excerpt_template,
                **extra_template_kwargs,
            )
            for i, text in enumerate(text_chunks)
        ]
    )


async def _create_excerpt(
    document: Document,
    text: str,
    excerpt_template: Template,
    **extra_template_kwargs,
) -> Document:
    keywords = extract_keywords(text)

    excerpt_text = await excerpt_template.render_async(
        document=document,
        excerpt_text=text,
        keywords=", ".join(keywords),
        **extra_template_kwargs,
    )
    return Document(
        parent_document_id=document.id,
        text=excerpt_text,
        keywords=keywords,
        metadata=document.metadata if document.metadata else {},
        tokens=count_tokens(excerpt_text),
    )


def get_distinct_documents(documents: Iterable["Document"]) -> Iterable["Document"]:
    """Return a list of distinct documents."""
    return distinct(documents, key=lambda doc: doc.hash)



"""Module for logging utilities."""

import logging
from functools import lru_cache, partial

from rich.logging import RichHandler
from rich.markup import escape

import raggy


class RaggyLogger(logging.Logger):
    """A subclass of the standard library `logging.Logger` class that adds methods for
    logging with styles and key-value pairs.
    """

    def __init__(self, name: str, level: int = logging.NOTSET) -> None:
        super().__init__(name, level)

    def debug_style(self, message: str, style: str | None = None) -> None:
        ...

    def info_style(self, message: str, style: str | None = None) -> None:
        ...

    def warning_style(self, message: str, style: str | None = None) -> None:
        ...

    def error_style(self, message: str, style: str | None = None) -> None:
        ...

    def critical_style(self, message: str, style: str | None = None) -> None:
        ...

    def debug_kv(
        self,
        key: str,
        value: str,
        key_style: str = "green",
        value_style: str = "default on default",
        delimiter: str = ": ",
    ) -> None:
        ...

    def info_kv(
        self,
        key: str,
        value: str,
        key_style: str = "blue",
        value_style: str = "default on default",
        delimiter: str = ": ",
    ) -> None:
        ...

    def warning_kv(
        self,
        key: str,
        value: str,
        key_style: str = "yellow",
        value_style: str = "default on default",
        delimiter: str = ": ",
    ) -> None:
        ...

    def error_kv(
        self,
        key: str,
        value: str,
        key_style: str = "red",
        value_style: str = "default on default",
        delimiter: str = ": ",
    ) -> None:
        ...

    def critical_kv(
        self,
        key: str,
        value: str,
        key_style: str = "red",
        value_style: str = "default on default",
        delimiter: str = ": ",
    ) -> None:
        ...


logging.setLoggerClass(RaggyLogger)


@lru_cache()
def get_logger(
    name: str | None = None,
) -> RaggyLogger:
    """
    Retrieves a logger with the given name, or the root logger if no name is given.

    Args:
        name: The name of the logger to retrieve.

    Returns:
        The logger with the given name, or the root logger if no name is given.

    Example:
        Basic Usage of `get_logger`
        ```python
        from raggy.utilities.logging import get_logger

        logger = get_logger("raggy.test")
        logger.info("This is a test") # Output: raggy.test: This is a test

        debug_logger = get_logger("raggy.debug")
        debug_logger.debug_kv("TITLE", "log message", "green")
        ```
    """
    parent_logger = logging.getLogger("raggy")

    if name:
        # Append the name if given but allow explicit full names e.g. "raggy.test"
        # should not become "raggy.raggy.test"
        if not name.startswith(parent_logger.name + "."):
            logger = parent_logger.getChild(name)
        else:
            logger = logging.getLogger(name)
    else:
        logger = parent_logger

    add_logging_methods(logger)
    return logger  # type: ignore


def setup_logging(level: str | None = None) -> None:
    logger: RaggyLogger = get_logger()

    if level is not None:
        logger.setLevel(level)
    else:
        logger.setLevel(raggy.settings.log_level)

    logger.handlers.clear()

    handler = RichHandler(rich_tracebacks=True, markup=False)
    formatter = logging.Formatter("%(name)s: %(message)s")
    handler.setFormatter(formatter)

    logger.addHandler(handler)
    logger.propagate = False


def add_logging_methods(logger: logging.Logger) -> None:
    def log_style(level: int, message: str, style: str | None = None):
        if not style:
            style = "default on default"
        message = f"[{style}]{escape(str(message))}[/]"
        logger.log(level, message, extra={"markup": True})

    def log_kv(
        level: int,
        key: str,
        value: str,
        key_style: str = "default on default",
        value_style: str = "default on default",
        delimiter: str = ": ",
    ):
        logger.log(
            level,
            f"[{key_style}]{escape(str(key))}{delimiter}[/][{value_style}]{escape(str(value))}[/]",
            extra={"markup": True},
        )

    setattr(logger, "debug_style", partial(log_style, logging.DEBUG))
    setattr(logger, "info_style", partial(log_style, logging.INFO))
    setattr(logger, "warning_style", partial(log_style, logging.WARNING))
    setattr(logger, "error_style", partial(log_style, logging.ERROR))
    setattr(logger, "critical_style", partial(log_style, logging.CRITICAL))

    setattr(logger, "debug_kv", partial(log_kv, logging.DEBUG))
    setattr(logger, "info_kv", partial(log_kv, logging.INFO))
    setattr(logger, "warning_kv", partial(log_kv, logging.WARNING))
    setattr(logger, "error_kv", partial(log_kv, logging.ERROR))
    setattr(logger, "critical_kv", partial(log_kv, logging.CRITICAL))



import re
from functools import lru_cache
from typing import (
    TYPE_CHECKING,
    TypeVar,
)

import tiktoken
import xxhash

import raggy

T = TypeVar("T")

if TYPE_CHECKING:
    pass


def rm_html_comments(text: str) -> str:
    return re.sub(r"<!--.*?-->", "", text, flags=re.DOTALL)


def rm_text_after(text: str, substring: str) -> str:
    return (
        text[: start + len(substring)]
        if (start := text.find(substring)) != -1
        else text
    )


def extract_keywords(text: str) -> list[str]:
    """Extract keywords from the given text using the yake library.

    Args:
        text: The text to extract keywords from.

    Returns:
        list[str]: The keywords extracted from the text.

    Raises:
        ImportError: If yake is not installed.

    Example:
        Extract keywords from a text:
        ```python
        from raggy.utilities.text import extract_keywords

        text = "This is a sample text from which we will extract keywords."
        keywords = extract_keywords(text)
        print(keywords) # ['keywords', 'sample', 'text', 'extract']
        ```
    """
    try:
        import yake
    except ImportError:
        raise ImportError(
            "yake is required for keyword extraction. Please install it with"
            " `pip install `raggy[rag]` or `pip install yake`."
        )

    kw = yake.KeywordExtractor(
        lan="en",
        n=1,
        dedupLim=0.9,
        dedupFunc="seqm",
        windowsSize=1,
        top=10,
        features=None,
    )

    return [k[0] for k in kw.extract_keywords(text)]


@lru_cache(maxsize=2048)
def hash_text(*text: str) -> str:
    """Hash the given text using the xxhash algorithm.

    Args:
        text: The text to hash.

    Returns:
        str: The hash of the text.

    Example:
        Hash a single text:
        ```python
        from raggy.utilities.text import hash_text

        text = "This is a sample text."
        hash_ = hash_text(text)
        print(hash_) # 4a2db845d20188ce069196726a065a09
        ```
    """
    bs = [t.encode() if not isinstance(t, bytes) else t for t in text]
    return xxhash.xxh3_128_hexdigest(b"".join(bs))


def get_encoding_for_model(model: str | None = None) -> tiktoken.Encoding:
    """Get the `tiktoken` encoding for the specified model.

    Args:
        model: The model to get the encoding for. If not provided, the default
            chat completions model is used (as specified in `raggy.settings`).
            If an invalid model is provided, 'gpt-3.5-turbo' is used.

    Returns:
        tiktoken.Encoding: The encoding for the specified model.

    Example:
        Get the encoding for the default chat completions model:
        ```python
        from raggy.utilities.text import get_encoding_for_model

        encoding = get_encoding_for_model() # 'gpt-3.5-turbo' by default
        ```
    """
    if model is None:
        model = raggy.settings.openai_chat_completions_model
    try:
        return tiktoken.encoding_for_model(model)
    except (KeyError, ValueError):
        return tiktoken.encoding_for_model("gpt-3.5-turbo")


def tokenize(text: str, model: str | None = None) -> list[int]:
    """
    Tokenizes the given text using the specified model.

    Args:
        text: The text to tokenize.
        model: The model to use for tokenization. If not provided,
            the default model is used.

    Returns:
        list[int]: The tokenized text as a list of integers.
    """
    return get_encoding_for_model(model).encode(text)


def detokenize(tokens: list[int], model: str | None = None) -> str:
    """
    Detokenizes the given tokens using the specified model.

    Args:
        tokens: The tokens to detokenize.
        model: The model to use for detokenization. If not provided,
            the default model is used.

    Returns:
        str: The detokenized text.
    """
    return get_encoding_for_model(model).decode(tokens)


def count_tokens(text: str, model: str | None = None) -> int:
    """
    Counts the number of tokens in the given text using the specified model.

    Args:
        text: The text to count tokens in.
        model: The model to use for token counting. If not provided,
            the default model is used.

    Returns:
        int: The number of tokens in the text.
    """
    return len(tokenize(text, model=model))


def slice_tokens(text: str, n_tokens: int) -> str:
    """Slices the given text to the specified number of tokens.

    Args:
        text: The text to slice.
        n_tokens: The number of tokens to slice the text to.

    Returns:
        str: The sliced text.

    Example:
        Slice a text to the first 50 tokens:
        ```python
        from raggy.utilities.text import slice_tokens

        text = "This is a sample text."*100
        sliced_text = slice_tokens(text, 5)
        print(sliced_text) # 'This is a sample text.'
        ```
    """
    return detokenize(tokenize(text)[:n_tokens])


def split_text(
    text: str,
    chunk_size: int,
    chunk_overlap: float | None = None,
    last_chunk_threshold: float | None = None,
) -> list[str]:
    """
    Split a text into a list of strings. Chunks are split by tokens.

    Args:
        text: The text to split.
        chunk_size: The number of tokens in each chunk.
        chunk_overlap: The fraction of overlap between chunks.
        last_chunk_threshold: If the last chunk is less than this fraction of
            the chunk_size, it will be added to the prior chunk

    Returns:
        list[str]: The list of chunks.

    Example:
        Split a text into chunks of 5 tokens with 10% overlap:
        ```python
        from raggy.utilities.text import split_text

        text = "This is a sample text."*3
        chunks = split_text(text, 5, 0.1)
        print(chunks) # ['This is a sample text', '.This is a sample text', '.This is a sample text.']
        ```
    """
    if chunk_overlap is None:
        chunk_overlap = 0.1
    if chunk_overlap < 0 or chunk_overlap > 1:
        raise ValueError("chunk_overlap must be between 0 and 1")
    if last_chunk_threshold is None:
        last_chunk_threshold = 0.25

    tokens = tokenize(text)

    chunks = []
    for i in range(0, len(tokens), chunk_size - int(chunk_overlap * chunk_size)):
        chunks.append((tokens[i : i + chunk_size], len(detokenize(tokens[:i]))))

    # if the last chunk is too small, merge it with the previous chunk
    if len(chunks) > 1 and len(chunks[-1][0]) < chunk_size * last_chunk_threshold:
        chunks[-2][0].extend(chunks.pop(-1)[0])

    return [detokenize(chunk) for chunk, _ in chunks]



from functools import partial
from typing import Any, Callable, TypeVar

import anyio

T = TypeVar("T")

RAGGY_THREAD_LIMITER: anyio.CapacityLimiter | None = None


def get_thread_limiter():
    global RAGGY_THREAD_LIMITER

    if RAGGY_THREAD_LIMITER is None:
        RAGGY_THREAD_LIMITER = anyio.CapacityLimiter(250)

    return RAGGY_THREAD_LIMITER


async def run_sync_in_worker_thread(
    __fn: Callable[..., T], *args: Any, **kwargs: Any
) -> T:
    """Runs a sync function in a new worker thread so that the main thread's event loop
    is not blocked

    Unlike the anyio function, this defaults to a cancellable thread and does not allow
    passing arguments to the anyio function so users can pass kwargs to their function.

    Note that cancellation of threads will not result in interrupted computation, the
    thread may continue running — the outcome will just be ignored.

    Args:
        __fn: The function to run in a worker thread
        *args: Positional arguments to pass to the function
        **kwargs: Keyword arguments to pass to the function

    Returns:
        The result of the function
    """
    call = partial(__fn, *args, **kwargs)
    return await anyio.to_thread.run_sync(
        call, cancellable=True, limiter=get_thread_limiter()
    )



import re
from typing import Iterable, Literal

try:
    from chromadb import Client, HttpClient
    from chromadb.api.models.Collection import Collection
    from chromadb.api.types import Include, QueryResult
except ImportError:
    raise ImportError(
        "You must have `chromadb` installed to use the Chroma vector store. "
        "Install it with `pip install 'raggy[chroma]'`."
    )

from raggy.documents import Document, get_distinct_documents
from raggy.utilities.asyncutils import run_sync_in_worker_thread
from raggy.utilities.embeddings import create_openai_embeddings
from raggy.utilities.text import slice_tokens
from raggy.vectorstores.base import Vectorstore


def get_client(client_type: Literal["base", "http"]) -> HttpClient:
    if client_type == "base":
        return Client()
    elif client_type == "http":
        return HttpClient()
    else:
        raise ValueError(f"Unknown client type: {client_type}")


class Chroma(Vectorstore):
    """A wrapper for chromadb.Client - used as an async context manager.

    Attributes:
        client_type: The type of client to use. Must be one of "base" or "http".
        collection_name: The name of the collection to use.

    Example:
        Query a collection:
        ```python
        from raggy.vectorstores.chroma import Chroma

        async with Chroma(collection_name="my-collection") as chroma:
            result = await chroma.query(query_texts=["Hello, world!"])
            print(result)
        ```

    """

    client_type: Literal["base", "http"] = "base"
    collection_name: str = "raggy"

    @property
    def collection(self) -> Collection:
        return get_client(self.client_type).get_or_create_collection(
            name=self.collection_name
        )

    async def delete(
        self,
        ids: list[str] | None = None,
        where: dict | None = None,
        where_document: Document | None = None,
    ):
        await run_sync_in_worker_thread(
            self.collection.delete,
            ids=ids,
            where=where,
            where_document=where_document,
        )

    async def add(self, documents: list[Document]) -> Iterable[Document]:
        documents = list(get_distinct_documents(documents))
        kwargs = dict(
            ids=[document.id for document in documents],
            documents=[document.text for document in documents],
            metadatas=[
                document.metadata.model_dump(exclude_none=True) or None
                for document in documents
            ],
            embeddings=await create_openai_embeddings(
                [document.text for document in documents]
            ),
        )

        await run_sync_in_worker_thread(self.collection.add, **kwargs)

        get_result = await run_sync_in_worker_thread(
            self.collection.get, ids=kwargs["ids"]
        )

        return get_result.get("documents")

    async def query(
        self,
        query_embeddings: list[list[float]] | None = None,
        query_texts: list[str] | None = None,
        n_results: int = 10,
        where: dict | None = None,
        where_document: dict | None = None,
        include: "Include" = ["metadatas"],
        **kwargs,
    ) -> "QueryResult":
        return await run_sync_in_worker_thread(
            self.collection.query,
            query_embeddings=query_embeddings,
            query_texts=query_texts,
            n_results=n_results,
            where=where,
            where_document=where_document,
            include=include,
            **kwargs,
        )

    async def count(self) -> int:
        return await run_sync_in_worker_thread(self.collection.count)

    async def upsert(self, documents: list[Document]):
        documents = list(get_distinct_documents(documents))
        kwargs = dict(
            ids=[document.id for document in documents],
            documents=[document.text for document in documents],
            metadatas=[
                document.metadata.model_dump(exclude_none=True) or None
                for document in documents
            ],
            embeddings=await create_openai_embeddings(
                [document.text for document in documents]
            ),
        )
        await run_sync_in_worker_thread(self.collection.upsert, **kwargs)

        get_result = await run_sync_in_worker_thread(
            self.collection.get, ids=kwargs["ids"]
        )

        return get_result.get("documents")

    async def reset_collection(self):
        client = get_client(self.client_type)
        try:
            await run_sync_in_worker_thread(
                client.delete_collection, self.collection_name
            )
        except ValueError:
            self.logger.warning_kv(
                "Collection not found",
                f"Creating a new collection {self.collection_name!r}",
            )
        await run_sync_in_worker_thread(client.create_collection, self.collection_name)

    def ok(self) -> bool:
        try:
            version = self.client.get_version()
        except Exception as e:
            self.logger.error_kv("Connection error", f"Cannot connect to Chroma: {e}")
        if re.match(r"^\d+\.\d+\.\d+$", version):
            self.logger.debug_kv("OK", f"Connected to Chroma {version!r}")
            return True
        return False


async def query_collection(
    query_text: str,
    query_embedding: list[float] | None = None,
    collection_name: str = "raggy",
    top_k: int = 10,
    where: dict | None = None,
    where_document: dict | None = None,
    max_tokens: int = 500,
) -> str:
    """Query a Chroma collection.

    Args:
        query_text: The text to query for.
        filters: Filters to apply to the query.
        collection: The collection to query.
        top_k: The number of results to return.

    Example:
        Basic query of a collection:
        ```python
        from raggy.vectorstores.chroma import query_collection

        print(await query_collection("How to create a flow in Prefect?"))
        ```
    """
    async with Chroma(collection_name=collection_name) as chroma:
        query_embedding = query_embedding or await create_openai_embeddings(query_text)

        query_result = await chroma.query(
            query_embeddings=[query_embedding],
            n_results=top_k,
            where=where,
            where_document=where_document,
            include=["documents"],
        )

        concatenated_result = "\n".join(
            doc for doc in query_result.get("documents", [])
        )

        return slice_tokens(concatenated_result, max_tokens)



"""Loaders for GitHub."""
import asyncio
import functools
import os
import re
from datetime import datetime
from pathlib import Path
from typing import Dict, List, Tuple

import aiofiles
import chardet
import httpx
from pydantic import BaseModel, Field, field_validator, model_validator

from raggy.documents import Document, document_to_excerpts
from raggy.loaders import Loader
from raggy.utilities.filesystem import OPEN_FILE_CONCURRENCY, multi_glob
from raggy.utilities.text import rm_html_comments, rm_text_after


async def read_file_with_chardet(file_path, errors="replace"):
    async with aiofiles.open(file_path, "rb") as f:
        content = await f.read()
        encoding = chardet.detect(content)["encoding"]

    async with aiofiles.open(file_path, "r", encoding=encoding, errors=errors) as f:
        text = await f.read()
    return text


class GitHubUser(BaseModel):
    login: str


class GitHubComment(BaseModel):
    body: str = Field(default="")
    user: GitHubUser = Field(default_factory=GitHubUser)


class GitHubLabel(BaseModel):
    name: str = Field(default="")


class GitHubIssue(BaseModel):
    created_at: datetime = Field(...)
    html_url: str = Field(...)
    number: int = Field(...)
    title: str = Field(default="")
    body: str | None = Field(default="")
    labels: List[GitHubLabel] = Field(default_factory=GitHubLabel)
    user: GitHubUser = Field(default_factory=GitHubUser)

    @field_validator("body")
    def validate_body(cls, v):
        if not v:
            return ""
        return v


class GitHubIssueLoader(Loader):
    """Loader for GitHub issues in a given repository.

    **Beware** the [GitHub API rate limit](https://docs.github.com/en/rest/overview/resources-in-the-rest-api#rate-limiting).

    Attributes:
        repo: The GitHub repository in the format 'owner/repo'.
        n_issues: The number of issues to load.
        include_comments: Whether to include comments in the issues.
        ignore_body_after: The text to ignore in the issue body.
        ignore_users: A list of users to ignore.
        use_GH_token: Whether to use the `GITHUB_TOKEN` environment variable for authentication (recommended).
    """

    source_type: str = "github issue"

    repo: str = Field(...)
    n_issues: int = Field(default=50)

    include_comments: bool = Field(default=False)
    ignore_body_after: str = Field(default="### Checklist")
    ignore_users: List[str] = Field(default_factory=list)
    use_GH_token: bool = Field(default=False)

    request_headers: Dict[str, str] = Field(default_factory=dict)

    @model_validator(mode="after")
    def auth_headers(self):
        self.request_headers.update({"Accept": "application/vnd.github.v3+json"})
        if self.use_GH_token and (token := os.getenv("GITHUB_TOKEN")):
            self.request_headers["Authorization"] = f"Bearer {token}"

    @staticmethod
    @functools.lru_cache(maxsize=2048)
    async def _get_issue_comments(
        repo: str,
        request_header_items: Tuple[Tuple[str, str]],
        issue_number: int,
        per_page: int = 100,
    ) -> List[GitHubComment]:
        """
        Get a list of all comments for the given issue.

        Returns:
            A list of dictionaries, each representing a comment.
        """
        url = f"https://api.github.com/repos/{repo}/issues/{issue_number}/comments"
        comments = []
        page = 1
        async with httpx.AsyncClient() as client:
            while True:
                response = await client.get(
                    url=url,
                    headers=dict(request_header_items),
                    params={"per_page": per_page, "page": page},
                )
                response.raise_for_status()
                if not (new_comments := response.json()):
                    break
                comments.extend([GitHubComment(**comment) for comment in new_comments])
                page += 1
            return comments

    async def _get_issues(self, per_page: int = 100) -> List[GitHubIssue]:
        """
        Get a list of all issues for the given repository.

        per_page: The number of issues to request per page.

        Returns:
            A list of `GitHubIssue` objects, each representing an issue.
        """  # noqa: E501
        url = f"https://api.github.com/repos/{self.repo}/issues"
        issues: List[GitHubIssue] = []
        page = 1
        async with httpx.AsyncClient() as client:
            while len(issues) < self.n_issues:
                remaining = self.n_issues - len(issues)
                response = await client.get(
                    url=url,
                    headers=self.request_headers,
                    params={
                        "per_page": min(remaining, per_page),
                        "page": page,
                        "include": "comments",
                    },
                )
                response.raise_for_status()
                if not (new_issues := response.json()):
                    break
                issues.extend([GitHubIssue(**issue) for issue in new_issues])
                page += 1
            return issues

    async def load(self) -> list[Document]:
        """
        Load all issues for the given repository.

        Returns:
            A list of `Document` objects, each representing an issue.
        """
        documents = []
        for issue in await self._get_issues():
            self.logger.debug(f"Found {issue.title!r}")
            clean_issue_body = rm_text_after(
                rm_html_comments(issue.body or ""), self.ignore_body_after
            )
            text = f"\n\n##**{issue.title}:**\n{clean_issue_body}\n\n"
            if self.include_comments:
                for (
                    comment
                ) in await self._get_issue_comments(  # hashable headers for lru_cache
                    self.repo, tuple(self.request_headers.items()), issue.number
                ):
                    if comment.user.login not in self.ignore_users:
                        text += f"**[{comment.user.login}]**: {comment.body}\n\n"
            metadata = dict(
                source=self.source_type,
                link=issue.html_url,
                title=issue.title,
                labels=", ".join([label.name for label in issue.labels]),
                created_at=issue.created_at.timestamp(),
            )
            documents.extend(
                await document_to_excerpts(
                    Document(
                        text=text,
                        metadata=metadata,
                    )
                )
            )
        return documents


class GitHubRepoLoader(Loader):
    """Loader for files on GitHub that match a glob pattern.


    Attributes:
        repo: The GitHub repository in the format 'owner/repo'.
        include_globs: A list of glob patterns to include.
        exclude_globs: A list of glob patterns to exclude.

    Raises:
        ValueError: If the repository is not in the format 'owner/repo'.

    Example:
        Load all files from the `prefecthq/prefect`
        ```python
        from raggy.loaders.github import GitHubRepoLoader

        loader = GitHubRepoLoader(repo="prefecthq/prefect")

        documents = await loader.load()
        print(documents)
        ```
    """

    source_type: str = "github source code"

    repo: str = Field(...)
    include_globs: list[str] = Field(default=None)
    exclude_globs: list[str] = Field(default=None)

    @field_validator("repo")
    def validate_repo(cls, v):
        if not re.match(r"^[^/\s]+/[^/\s]+$", v):
            raise ValueError(
                "Must provide a GitHub repository in the format 'owner/repo'"
            )
        return f"https://github.com/{v}.git"

    async def load(self) -> list[Document]:
        """Load files from GitHub that match the glob pattern."""
        async with OPEN_FILE_CONCURRENCY:
            async with aiofiles.tempfile.TemporaryDirectory(suffix="_raggy") as tmp_dir:
                process = await asyncio.create_subprocess_exec(
                    *["git", "clone", "--depth", "1", self.repo, tmp_dir],
                    stdout=asyncio.subprocess.PIPE,
                    stderr=asyncio.subprocess.PIPE,
                )

                if (await process.wait()) != 0:
                    raise OSError(
                        f"Failed to clone repository:\n {await process.stderr.read() if process.stderr else ''}"
                    )

                self.logger.debug(
                    f"{await process.stdout.read() if process.stdout else ''}"
                )

                # Read the contents of each file that matches the glob pattern
                documents = []

                for file in multi_glob(tmp_dir, self.include_globs, self.exclude_globs):
                    self.logger.debug(f"Loading file: {file!r}")

                    metadata = dict(
                        source=self.source_type,
                        link="/".join(
                            [
                                self.repo.replace(".git", ""),
                                "tree/main",
                                str(file),
                            ]
                        ),
                        title=file.name,
                        filename=file.name,
                    )
                    documents.extend(
                        await document_to_excerpts(
                            Document(
                                text=await read_file_with_chardet(Path(tmp_dir) / file),
                                metadata=metadata,
                            )
                        )
                    )
                return documents



import itertools
from typing import Any, Callable, Generator, Iterable, TypeVar

T = TypeVar("T")


def distinct(
    iterable: Iterable[T],
    key: Callable[[T], Any] = (lambda i: i),
) -> Generator[T, None, None]:
    """Yield distinct items from an iterable.

    Args:
        iterable: The iterable to filter
        key: A function to compute a key for each item

    Yields:
        Distinct items from the iterable

    Example:
        Dedupe a list of Pydantic models by a key:
        ```python
        from pydantic import BaseModel
        from raggy.utilities.collections import distinct

        class MyModel(BaseModel):
            id: int
            name: str

        items = [
            MyModel(id=1, name="foo"),
            MyModel(id=2, name="bar"),
            MyModel(id=1, name="baz"),
        ]

        deduped = list(distinct(items, key=lambda i: i.id))

        assert deduped == [
            MyModel(id=1, name="foo"),
            MyModel(id=2, name="bar"),
        ]
        ```
    """
    seen: set = set()
    for item in iterable:
        if key(item) in seen:
            continue
        seen.add(key(item))
        yield item


def batched(
    iterable: Iterable[T], size: int, size_fn: Callable[[T], int] | None = None
) -> Generator[tuple[T, ...], None, None]:
    """Yield batches of items from an iterable.

    If size_fn is not provided, then the batch size will be determined by the
    number of items in the batch.

    If size_fn is provided, then it will be used
    to compute the batch size. Note that if a single item is larger than the
    batch size, it will be returned as a batch of its own.

    Args:
        iterable: The iterable to batch
        size: The size of the batch
        size_fn: A function to compute the size of an item in the iterable

    Yields:
        A batch of items from the iterable

    Example:
        Batch a list of strings by the number of characters:
        ```python
        from raggy.utilities.collections import batched

        items = [
            "foo",
            "bar",
            "baz",
            "qux",
            "quux",
            "corge",
            "grault",
            "garply",
            "waldo",
            "fred",
            "plugh",
            "xyzzy",
            "thud",
        ]

        batches = list(batched(items, size=10, size_fn=len))

        assert batches == [
            ('foo', 'bar', 'baz'),
            ('qux', 'quux'),
            ('corge',),
            ('grault',),
            ('garply',),
            ('waldo', 'fred'),
            ('plugh', 'xyzzy'),
            ('thud',)
        ]
        ```
    """
    if size_fn is None:
        it = iter(iterable)
        while True:
            batch_tuple = tuple(itertools.islice(it, size))
            if not batch_tuple:
                break
            yield batch_tuple
    else:
        batch_list: list[T] = []
        batch_size = 0
        for item in iterable:
            item_size = size_fn(item)
            if batch_size + item_size > size and batch_list:
                yield tuple(batch_list)
                batch_list = []
                batch_size = 0
            batch_list.append(item)
            batch_size += item_size
        if batch_list:
            yield tuple(batch_list)



import asyncio
from abc import ABC, abstractmethod

from pydantic import BaseModel, ConfigDict

from raggy.documents import Document
from raggy.utilities.collections import batched
from raggy.utilities.logging import get_logger


class Loader(BaseModel, ABC):
    """A base class for loaders."""

    model_config = ConfigDict(arbitrary_types_allowed=True, extra="forbid")

    @abstractmethod
    async def load(self) -> list[Document]:
        pass

    @property
    def logger(self):
        return get_logger(self.__class__.__name__)


class MultiLoader(Loader):
    """A loader that loads from multiple loaders.

    Attributes:
        loaders: The loaders to load from.

    Examples:
        Basic Usage of `MultiLoader`
        ```python
        from raggy.loaders.base import MultiLoader
        from raggy.loaders.github import GitHubRepoLoader

        loader = MultiLoader(
            loaders=[
                GitHubRepoLoader(repo="prefecthq/prefect"),
                GitHubRepoLoader(repo="prefecthq/marvin"),
            ]
        )

        documents = await loader.load() # all (chunked) files from both repos
        print(documents)
        ```

    """

    loaders: list[Loader]

    async def load(self, batch_size: int = 5) -> list[Document]:
        return [
            doc
            for batch in batched(self.loaders, batch_size)
            for docs in await asyncio.gather(*(loader.load() for loader in batch))
            for doc in docs
        ]



import asyncio
import math
import os
from pathlib import Path


def multi_glob(
    directory: str | None = None,
    keep_globs: list[str] | None = None,
    drop_globs: list[str] | None = None,
) -> list[Path]:
    """
    Return a list of all files in the given directory that match the
    patterns in keep_globs and do not match the patterns in drop_globs.
    The patterns are defined using glob syntax.

    Args:
        directory: The directory to search in. If not provided, the current
            working directory is used.
        keep_globs: A list of glob patterns to keep.
        drop_globs: A list of glob patterns to drop.

    Returns:
        A list of Path objects representing the files that match the given
        patterns.

    Raises:
        ValueError: If the directory does not exist.

    Example:
        Get all files (except those in the .git directory) in the current directory
        ```python
        from raggy.utilities.filesystem import multi_glob

        files = multi_glob() # .git files are excluded by default (unless drop_globs is set)
        ```

        Get all python files in the current directory
        ```python
        from raggy.utilities.filesystem import multi_glob

        files = multi_glob(keep_globs=["**/*.py"])
        ```

        Get all files except those in any `__pycache__` directories
        ```python
        from raggy.utilities.filesystem import multi_glob

        files = multi_glob(drop_globs=["**/__pycache__/**/*"])
        ```
    """
    keep_globs = keep_globs or ["**/*"]
    drop_globs = drop_globs or [".git/**/*"]
    directory_path = Path(directory) if directory else Path.cwd()

    if not directory_path.is_dir():
        raise ValueError(f"'{directory}' is not a directory.")

    def files_from_globs(globs: list[str]) -> set[Path]:
        return {
            file
            for pattern in globs
            for file in directory_path.glob(pattern)
            if file.is_file()
        }

    matching_files = files_from_globs(keep_globs) - files_from_globs(drop_globs)
    return [file.relative_to(directory_path) for file in matching_files]


def get_open_file_limit() -> int:
    """Get the maximum number of open files allowed for the current process.

    Returns:
        The maximum number of open files allowed for the current process.
    """
    try:
        if os.name == "nt":
            import ctypes

            return ctypes.cdll.ucrtbase._getmaxstdio()
        else:
            import resource

            soft_limit, _ = resource.getrlimit(resource.RLIMIT_NOFILE)
            return soft_limit
    except Exception:
        return 200


OPEN_FILE_CONCURRENCY = asyncio.Semaphore(math.floor(get_open_file_limit() / 2))



from contextlib import asynccontextmanager
from tempfile import NamedTemporaryFile
from typing import List
from urllib.parse import urlparse

import httpx

try:
    import pypdf
except ModuleNotFoundError:
    raise ImportError(
        "The PDF loader requires the pypdf package. "
        "Install it with `pip install 'raggy[pdf]'`."
    )

from raggy.documents import Document, document_to_excerpts
from raggy.loaders.base import Loader


async def download_url_content(url: str) -> bytes:
    async with httpx.AsyncClient() as client:
        response = await client.get(url)
        return response.content


def is_valid_url(url):
    parsed = urlparse(url)
    return bool(parsed.netloc) and bool(parsed.scheme)


class PDFLoader(Loader):
    """A loader for PDF files.

    Attributes:
        file_path: The path to the PDF file or a URL to download the PDF from.

    Examples:
        Load a PDF file from a local path:
        ```python
        from raggy.loaders.pdf import PDFLoader

        loader = PDFLoader(file_path="path/to/file.pdf")
        documents = await loader.load()
        print(documents)
        ```

        Load a PDF file from a URL:
        ```python
        from raggy.loaders.pdf import PDFLoader

        loader = PDFLoader(file_path="https://www.w3.org/WAI/ER/tests/xhtml/testfiles/resources/pdf/dummy.pdf")
        documents = await loader.load()
        print(documents)
        ```

    """

    file_path: str

    @asynccontextmanager
    async def open_pdf_file(self, file_path: str):
        if is_valid_url(file_path):
            raw_pdf_content = await download_url_content(file_path)
            with NamedTemporaryFile() as temp_file:
                temp_file.write(raw_pdf_content)
                temp_file.flush()
                yield temp_file
        else:
            with open(file_path, "rb") as pdf_file_obj:
                yield pdf_file_obj

    async def load(self) -> List[Document]:
        async with self.open_pdf_file(self.file_path) as pdf_file_obj:
            pdf_reader = pypdf.PdfReader(pdf_file_obj)
            return [
                document
                for i, page in enumerate(pdf_reader.pages)
                for document in await document_to_excerpts(
                    Document(
                        text=page.extract_text(),
                        metadata={"page": i + 1, "file_path": self.file_path},
                    )
                )
            ]



from openai import APIConnectionError, AsyncOpenAI
from openai.types import CreateEmbeddingResponse
from tenacity import retry, retry_if_exception_type, stop_after_attempt, wait_fixed

import raggy


@retry(
    retry=retry_if_exception_type(APIConnectionError),
    stop=stop_after_attempt(3),
    wait=wait_fixed(2),
)
async def create_openai_embeddings(
    input_: str | list[str],
    timeout: int = 60,
    model: str = raggy.settings.openai_embeddings_model,
) -> list[float] | list[list[float]]:
    """Create OpenAI embeddings for a list of texts.

    Args:
        input_: The input text or list of texts to embed
        timeout: The maximum time to wait for the request to complete
        model: The model to use for the embeddings. Defaults to the value
            of `raggy.settings.openai_embeddings_model`, which is "text-embedding-3-small"
            by default

    Returns:
        The embeddings for the input text or list of texts

    Raises:
        TypeError: If input_ is not a str or a list of str

    Examples:
        Create an embedding for a single text:
        ```python
        from raggy.utilities.embeddings import create_openai_embeddings

        embedding = await create_openai_embeddings("Hello, world!")
        ```

        Create embeddings for a list of texts:
        ```python
        from raggy.utilities.embeddings import create_openai_embeddings

        embeddings = await create_openai_embeddings([
            "Hello, world!",
            "Goodbye, world!",
        ])
        ```
    """

    if isinstance(input_, str):
        input_ = [input_]
    elif not isinstance(input_, list):
        raise TypeError(
            f"Expected input to be a str or a list of str, got {type(input_).__name__}."
        )

    embedding: CreateEmbeddingResponse = await AsyncOpenAI().embeddings.create(
        input=input_, model=model, timeout=timeout
    )

    if len(embedding.data) == 1:
        return embedding.data[0].embedding

    return [data.embedding for data in embedding.data]
